//==============================================================================
//! \file      ddx-graphic_font.h
//! \brief     FIXME!
//!
//! \author    David Odin <david.odin@forma3dev.fr>
//! \date      Created on 2009-08-01
///
//! Modifications: $Id: ddx-graphic_font.h 1182 2012-11-06 19:53:44Z dindinx $
///
//! \note (c) Forma3Dev - forma3dev.fr
//==============================================================================

#ifndef DDX_GRAPHIC__FONT_H
#define DDX_GRAPHIC__FONT_H

#include <vector>
#include <map>

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H

#include "shader.h"
#include "singleton.h"
#include "vec2.h"

//#include "ddx-graphic_shader.h"

// overall manager, main api entry-points
class FontManager
{
  struct FontFaceId
  {
    FontFaceId(const std::string &name, int size, bool from_serial = false);
    std::string font_name;
    int         font_size;
    bool        from_serial;
    bool operator<(const FontFaceId &id)const
    {
      if ( font_name <  id.font_name ||
           (font_name == id.font_name && font_size < id.font_size) ||
           (font_name == id.font_name && font_size ==id.font_size && from_serial == false && id.from_serial == true)
         )
        return true;
      return false;
    }
  };
  // managing one font with one size but using the FontManager vao
  class FontFace
  {
    struct Glyph
    {
      int         width;          //!<
      int         height;         //!<
      unsigned    texture_index;  //!<
      int         index;          //!<
      vec2        uv[4];          //!<
      int         left;           //!<
      int         top;            //!<
      int         advance_x;      //!<
      int         advance_y;      //!<
      unsigned char *data;
    };

    FT_Face              freetype_face;
    FT_Library          &freetype_library;
    std::map<int, Glyph> glyphes;
    unsigned             x_off;
    unsigned             y_off;
    int                  y_max;
    std::vector<unsigned>  textures;
    std::vector<std::vector<vec2>> aux_buffers;
    std::vector<FT_Byte> font_content;

  public:
    FontFace(FT_Library &library, const std::string &filename, int size, int ppi);
    FontFace(FT_Library &library, const char *serial_buffer, const std::string &filename, int size, int ppi);
    Glyph get_glyph(const char *&text);
    void draw_text(int x, int y, const char *text);
    void get_text_size(const char *text, int font_height, int ppi,
                       int &string_width, int &string_height,
                       int &string_ascent);
    void get_glyph_face(const char *text, int &width, int &height, const unsigned char *&data);
  };
  std::map<FontFaceId, FontFace*> font_faces;
  FT_Library           freetype_library;
  unsigned              vao;
  unsigned              vbo;
  unsigned              vbo_indices;
  Shader               *program;
  std::string          current_font_name;
  int                  current_font_size;
  int                  current_ppi;
  FontFace            *current_font_face;

  friend struct CreateUsingNew<FontManager>;
  FontManager();
  ~FontManager();
  void init_common();
public:
  static FontManager &instance(){return SingletonHolder<FontManager>::instance();}
#define FONT_MANAGER FontManager::instance()
  void init();
  void exit();
  void set_color(const vec4 &color);
  void set_screen_size(const vec2 &size);
  void set_font(const std::string &filename, int size = 10);
  void set_font_size(int size);
  void set_ppi(int ppi);

  // Proxy
  void draw_text(int x, int y, const char *text);
  void get_text_size(const char *text,
                     int &string_width, int &string_height,
                     int &string_ascent);
  void get_glyph_face(const char *text, int &width, int &height, const unsigned char *&data);
};
#endif
