// Fichier principal de l'utilisation du réseau de neurones

#include <algorithm>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

#include <cstdio>
#include <cstdlib>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>

#define GLEW_STATIC 1
#include <GL/glew.h>
#include <GL/freeglut.h>

#include "font-manager.h"
#include "gl-utils.h"
#include "reseau.h"
#include "shader.h"
#include "vec3.h"
#include "vec4.h"

struct Font
{
  std::string name;
  bool active;
  Font(const std::string &name=""):name(name),active(true){}
};

std::vector<Font> all_fonts;
std::vector<std::string> active_fonts;

int BRUIT = 20;
int EPSILON = 10;
bool normalized = true;
int taille_image = 24;
int min_font_size = 24;
int max_font_size = 24;

static int neurones_par_couche[3];

static int iteration;
static Reseau Net;
static int type; // Le type de forme qui est en train d'être apprise
static float donnees_entree[TAILLE_COUCHE_ENTREE];
static float donnees_sortie[TAILLE_COUCHE_SORTIE];
static int Success2;
static int Success3[100];

static std::vector<vec3> rectangles;
static GLuint rectangles_vao, rectangles_vbo;
static Shader *rectangles_shader;

static GLuint neurone_vao, neurone_vbos[2];
static GLuint neurone_texture;
static Shader *neurone_shader;

/*===========================================================================*\
 * init_opengl                                                               *
\*===========================================================================*/
static void init_opengl()
{
  std::string vertex_shader =
    "#version 330 core\n"
    "\n"
    "in vec2 position;\n"
    "in vec2 tex_coord;\n"
    "\n"
    "uniform mat4 projection;\n"
    "\n"
    "out vec2 v_tex_coord;\n"
    "\n"
    "void main(void)\n"
    "{\n"
    "  gl_Position = projection * vec4(position, 0.0, 1.0);\n"
    "  v_tex_coord = tex_coord;\n"
    "}\n";

  std::string fragment_shader =
    "#version 330 core\n"
    "\n"
    "uniform sampler2D tex;\n"
    "\n"
    "in vec2 v_tex_coord;\n"
    "out vec4 color;\n"
    "\n"
    "void main(void)\n"
    "{\n"
    "  float a = texture(tex, v_tex_coord).r;\n"
    "  color = vec4(a, a, a, 1.0);\n"
    "}\n";
  neurone_shader = new Shader();
  neurone_shader->attach_vertex(vertex_shader);
  neurone_shader->attach_fragment(fragment_shader);
  neurone_shader->set_attributes({"position", "tex_coord"});
  neurone_shader->link();
  neurone_shader->activate();
  matrix projection;
  projection.set_ortho(0,1000,63*6-15,0,-1,1);
  neurone_shader->set("projection", projection);
  neurone_shader->set("tex", 0);

  GLushort indices[] = { 0,1,2, 0,2,3 };
  glGenVertexArrays(1, &neurone_vao);                                                             PRINT_OPENGL_ERROR();
  glBindVertexArray(neurone_vao);                                                                 PRINT_OPENGL_ERROR();
  glGenBuffers(2, neurone_vbos);                                                                  PRINT_OPENGL_ERROR();
  glBindBuffer(GL_ARRAY_BUFFER, neurone_vbos[0]);                                                 PRINT_OPENGL_ERROR();
  glEnableVertexAttribArray(0);                                                                   PRINT_OPENGL_ERROR();
  glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 2 * sizeof(vec2), BUFFER_OFFSET(0));            PRINT_OPENGL_ERROR();
  glEnableVertexAttribArray(1);                                                                   PRINT_OPENGL_ERROR();
  glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 2 * sizeof(vec2), BUFFER_OFFSET(sizeof(vec2))); PRINT_OPENGL_ERROR();
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, neurone_vbos[1]);                                         PRINT_OPENGL_ERROR();
  glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof indices, indices, GL_STATIC_DRAW);                 PRINT_OPENGL_ERROR();
  glBindVertexArray(0);                                                                           PRINT_OPENGL_ERROR();

  glGenTextures(1, &neurone_texture);                                  PRINT_OPENGL_ERROR();
  glBindTexture(GL_TEXTURE_2D, neurone_texture);                       PRINT_OPENGL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE); PRINT_OPENGL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE); PRINT_OPENGL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);   PRINT_OPENGL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);   PRINT_OPENGL_ERROR();
}

/*===========================================================================*\
 * neurone_set_texture                                                       *
\*===========================================================================*/
static void neurone_set_texture(int width, int height, float *data)
{
  glActiveTexture(GL_TEXTURE0);                                                      PRINT_OPENGL_ERROR();
  glBindTexture(GL_TEXTURE_2D, neurone_texture);                                     PRINT_OPENGL_ERROR();
  glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, width, height, 0, GL_RED, GL_FLOAT, data); PRINT_OPENGL_ERROR();
}

/*===========================================================================*\
 * draw_neurone                                                              *
\*===========================================================================*/
static void draw_neurone(const vec2 &min_pos, const vec2 &max_pos, const vec2 &min_tex, const vec2 &max_tex)
{
  std::vector<vec2> vertices;
  vertices.push_back(vec2(min_pos.x, min_pos.y)); vertices.push_back(vec2(min_tex.x, min_tex.y));
  vertices.push_back(vec2(min_pos.x, max_pos.y)); vertices.push_back(vec2(min_tex.x, max_tex.y));
  vertices.push_back(vec2(max_pos.x, max_pos.y)); vertices.push_back(vec2(max_tex.x, max_tex.y));
  vertices.push_back(vec2(max_pos.x, min_pos.y)); vertices.push_back(vec2(max_tex.x, min_tex.y));

  glBindVertexArray(neurone_vao);                                                                  PRINT_OPENGL_ERROR();
  glBindBuffer(GL_ARRAY_BUFFER, neurone_vbos[0]);                                                  PRINT_OPENGL_ERROR();
  glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(vec2), vertices.data(), GL_DYNAMIC_DRAW); PRINT_OPENGL_ERROR();
  neurone_shader->activate();
  glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, nullptr);                                     PRINT_OPENGL_ERROR();
  glBindVertexArray(0);                                                                            PRINT_OPENGL_ERROR();
}
/*===========================================================================*\
 * rectangles_init                                                           *
\*===========================================================================*/
static void rectangles_init()
{
  std::string vertex_shader =
    "#version 330 core\n"
    "\n"
    "in vec3 position;\n"
    "\n"
    "uniform mat4 projection;\n"
    "\n"
    "out vec4 v_color;\n"
    "\n"
    "void main(void)\n"
    "{\n"
    "  gl_Position = projection * vec4(position.xy, 0.0, 1.0);\n"
    "  v_color = vec4(position.zzz, 1.0);\n"
    "}\n";

  std::string fragment_shader =
    "#version 330 core\n"
    "\n"
    "in vec4 v_color;\n"
    "out vec4 color;\n"
    "\n"
    "void main(void)\n"
    "{\n"
    "  color = v_color;\n"
    "}\n";

  rectangles_shader = new Shader();
  rectangles_shader->attach_vertex(vertex_shader);
  rectangles_shader->attach_fragment(fragment_shader);
  rectangles_shader->set_attributes({"position"});
  rectangles_shader->link();
  rectangles_shader->activate();
  matrix projection;
  projection.set_ortho(0,1000,63*6-15,0,-1,1);
  rectangles_shader->set("projection", projection);

  glGenVertexArrays(1, &rectangles_vao);                 PRINT_OPENGL_ERROR();
  glBindVertexArray(rectangles_vao);                     PRINT_OPENGL_ERROR();
  glGenBuffers(1, &rectangles_vbo);                      PRINT_OPENGL_ERROR();
  glBindBuffer(GL_ARRAY_BUFFER, rectangles_vbo);         PRINT_OPENGL_ERROR();
  glEnableVertexAttribArray(0);                          PRINT_OPENGL_ERROR();
  glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0); PRINT_OPENGL_ERROR();
  glBindVertexArray(0);                                  PRINT_OPENGL_ERROR();
}

/*===========================================================================*\
 * flush_rectangles                                                          *
\*===========================================================================*/
static void flush_rectangles()
{
  if (rectangles.empty()) return;
  glBindVertexArray(rectangles_vao);                                                                   PRINT_OPENGL_ERROR();
  glBindBuffer(GL_ARRAY_BUFFER, rectangles_vbo);                                                       PRINT_OPENGL_ERROR();
  glBufferData(GL_ARRAY_BUFFER, rectangles.size() * sizeof(vec3), rectangles.data(), GL_DYNAMIC_DRAW); PRINT_OPENGL_ERROR();
  rectangles_shader->activate();
  glDrawArrays(GL_TRIANGLES, 0, rectangles.size());                                                    PRINT_OPENGL_ERROR();
  glBindVertexArray(0);                                                                                PRINT_OPENGL_ERROR();
  rectangles.clear();
}

/*===========================================================================*\
 * Rectangle                                                                 *
 * dessine un rectangle entre deux points                                    *
\*===========================================================================*/
static void Rectangle(int x1, int y1, int x2, int y2, float gray_level)
{
  rectangles.push_back(vec3(x1, y1, gray_level));
  rectangles.push_back(vec3(x1, y2, gray_level));
  rectangles.push_back(vec3(x2, y2, gray_level));
  rectangles.push_back(vec3(x1, y1, gray_level));
  rectangles.push_back(vec3(x2, y2, gray_level));
  rectangles.push_back(vec3(x2, y1, gray_level));

  if (rectangles.size() > 1000)
    flush_rectangles();
}

/*===========================================================================*\
 * display_callback                                                          *
 * dessine un le réseau de neurones                                          *
\*===========================================================================*/
static void display_callback()
{
  char buffer[1000];
  glClear(GL_COLOR_BUFFER_BIT);                      PRINT_OPENGL_ERROR();
  glEnable(GL_BLEND);                                PRINT_OPENGL_ERROR();
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); PRINT_OPENGL_ERROR();
 
  // Remise en place de la police par défaut pour les affichages
  FONT_MANAGER.set_font("fonts/luxisr.ttf", 8);
  FONT_MANAGER.set_color(vec4(1, 1, 1, 1));

  // Affichage de statistiques générales
  sprintf(buffer, "Iteration: %d", iteration);
  FONT_MANAGER.draw_text(10, 40, buffer);

  sprintf(buffer, "Epsilon = %d %%", EPSILON);
  FONT_MANAGER.draw_text(10, 70, buffer);

  sprintf(buffer, "Bruit : %d %%", BRUIT);
  FONT_MANAGER.draw_text(10, 100, buffer);

  sprintf(buffer, "%% de reussite global : %6.2f", Success2 * 100.0 / iteration);
  FONT_MANAGER.draw_text(710, 30, buffer);

  sprintf(buffer, "%% de reussite instantane : %d", Success3[(iteration+1)%100]+1);
  FONT_MANAGER.draw_text(710, 15, buffer);

  sprintf(buffer, "Taille police : %d-%d", min_font_size, max_font_size);
  FONT_MANAGER.draw_text(550, 20, buffer);

  FONT_MANAGER.draw_text(900, 50, "Polices disponibles :");
  for (unsigned i = 0; i < all_fonts.size(); i++)
  {
    if (all_fonts[i].active)
      FONT_MANAGER.set_color(vec4(1, 1, 1, 1));
    else
      FONT_MANAGER.set_color(vec4(0.5, 0.5, 0.5, 1));
    FONT_MANAGER.draw_text(920, 70+20*i, all_fonts[i].name.c_str());
  }

  FONT_MANAGER.set_color(vec4(1, 1, 1, 1));
  // Affichage de l'image d'entrée
  FONT_MANAGER.draw_text(10, 140, "Image d'entree");
  std::vector<float> neurone;
  for (int u = 0; u < TAILLE_COUCHE_ENTREE; u++)
  {
    float gray_level = (Net.unit[0][u].etat + BRUIT / 200.0 ) / (1 + BRUIT / 100.0);
    neurone.push_back(gray_level);
  }
  neurone_set_texture(LARGEUR_IMAGE, HAUTEUR_IMAGE, neurone.data());
  draw_neurone(vec2(10, 150), vec2(90, 230), vec2(0, 0), vec2(1, 1));

  float min, max;
  { // Affichage des poids de la couche intermédiaire
    if (normalized)
    {
      min = 500000.0; max = -min;
      for (int i = 0; i < TAILLE_COUCHE_CACHEE; i++)
        for (unsigned j = 0; j < Net.unit[1][i].poids.size(); j++)
        {
          max = std::max(max, Net.unit[1][i].poids[j]);
          min = std::min(min, Net.unit[1][i].poids[j]);
        }  
    } else
    {
      min = 0; max = 1;
    }
    for (int i = 0; i < TAILLE_COUCHE_CACHEE; i++)
    {
      neurone.clear();
      for (int j = 0; j < TAILLE_COUCHE_ENTREE; j++)
      {
        float gray_level = (Net.unit[1][i].poids[j] - min) / (max - min);
        neurone.push_back(gray_level);
      }
      neurone_set_texture(LARGEUR_IMAGE, HAUTEUR_IMAGE, neurone.data());
      draw_neurone(vec2(100+(i%6)*63, (i/6)*63), vec2(100+(i%6)*63+56, (i/6)*63+56), vec2(0, 0), vec2(1, 1));
    }
  }

  { // Affichage des poids de la couche de sortie
    if (normalized)
    {
      min = 500000.0; max = -min;
      for (int i = 0; i < neurones_par_couche[2]; i++)
        for (int j = 0; j < TAILLE_COUCHE_CACHEE; j++)
        {
          max = std::max(max, Net.unit[2][i].poids[j]);
          min = std::min(min, Net.unit[2][i].poids[j]);
        }
    } else
    {
      min = 0; max = 1;
    }
    for (int i = 0; i < neurones_par_couche[2]; i++)
    {
      for (int j = 0; j < TAILLE_COUCHE_CACHEE; j++)
      {
        float gray_level = (Net.unit[2][i].poids[j] - min) / (max - min);
        Rectangle(100 + 63*6 + (j%6) * 6 + (i%6) * 50,      50 + (j/6) * 6 + (i/6) * 55,
                  100 + 63*6 + (j%6) * 6 + 12 + (i%6) * 50, 50 + (j/6) * 6 + 12 + (i/6) * 55, gray_level);
      }
      std::string letter("0");
      letter[0] = 'A' + i;
      FONT_MANAGER.draw_text(100 + 63 * 6 + (i % 6) * 50, 48 + (i / 6) * 55, letter.c_str());      
      FONT_MANAGER.draw_text(785 + (i / 13) * 60, 67 + (i % 13) * 20, letter.c_str());      
    }
  }
  {  // Affichage de l'état d'activation de la couche de sortie :
    for (int i = 0; i < neurones_par_couche[2]; i++)
    {
      Rectangle(800 + (i / 13) * 60,      50 + (i % 13) * 20,
                800 + (i / 13) * 60 + 20, 50 + (i % 13) * 20 + 20,
                Net.unit[2][i].etat);

      Rectangle(820 + (i / 13) * 60,      50 + (i % 13) * 20,
                820 + (i / 13) * 60 + 20, 50 + (i % 13) * 20 + 20, donnees_sortie[i]);
    }
  }
  flush_rectangles();
  glutSwapBuffers();
}

/*===========================================================================*\
 * idle_callback                                                             *
 * fonction appelée en boucle le plus vite possible                          *
\*===========================================================================*/
static void idle_callback()
{
  iteration++;

  // Remise à zéro de la couche d'entrée
  for (int i = 0; i < HAUTEUR_IMAGE; i++)
    for (int j = 0; j < LARGEUR_IMAGE; j++)
      donnees_entree[i * LARGEUR_IMAGE + j] = 0.0;
  // Remise à zéro de la couche de sortie
  for (int i = 0; i < TAILLE_COUCHE_SORTIE; i++)
    donnees_sortie[i] = 0.0;
  //
  type = rand() % TAILLE_COUCHE_SORTIE;
  donnees_sortie[type] = 1.0f;

  std::string text("0");
  text[0] = 'A' + type;

  int glyph_width, glyph_height;
  const unsigned char *data;
  // Choix de la police et de sa taille
  FONT_MANAGER.set_font(active_fonts[rand()%active_fonts.size()],
                        min_font_size + rand() % (max_font_size - min_font_size + 1));
  // rendu de la lettre dans data
  FONT_MANAGER.get_glyph_face(text.c_str(), glyph_width, glyph_height, data);

  // Décalage aléatoire de la lettre s'il y a la place.
  int x_off = 0;
  if (LARGEUR_IMAGE > glyph_width) x_off = rand()%(LARGEUR_IMAGE + 1 - glyph_width);
  int y_off = 0;
  if (HAUTEUR_IMAGE > glyph_height) y_off  = rand()%(HAUTEUR_IMAGE + 1 - glyph_height);

  // Remplissage de la couche d'entrée avec la lettre
  for (int i = 0; i < std::min(LARGEUR_IMAGE, glyph_width); i++)
    for (int j = 0; j < std::min(HAUTEUR_IMAGE, glyph_height); j++)
      donnees_entree[(j + y_off) * LARGEUR_IMAGE + i + x_off] = data[j * glyph_width + i] / 255.0;
  // Ajout de bruit
  for (int i = 0; i < HAUTEUR_IMAGE; i++)
    for (int j = 0; j < LARGEUR_IMAGE; j++) 
      donnees_entree[i * LARGEUR_IMAGE + j] += (rand() % 100) / 100.0 * (BRUIT / 100.0) - BRUIT / 200.0;

  // Propagation
  Net.copie_image_entree(donnees_entree);
  for (int couche = 1; couche < 3; couche++)
    Net.maj_activation(couche);

  // Retropropagation
  Net.retro_propagation(donnees_sortie);

  // On n'affiche qu'une fois sur mille
  if ((iteration % 1000) == 0) glutPostRedisplay();

  // Mise à jour des statistiques
  int resultat = Net.get_max_out_index();
  if (type == resultat)
  {
    Success2++;
    for (int j = 0; j < 100; j++)
      Success3[j]++;
  }
  Success3[iteration % 100] = 0;
}

/*===========================================================================*\
 * keyboard_callback                                                         *
\*===========================================================================*/
static void keyboard_callback(unsigned char key, int, int)
{
  switch (key)
  {
    case 27: exit(0);
    case 'b': BRUIT--; if (BRUIT < 0) BRUIT = 0; break;
    case 'B': BRUIT++; if (BRUIT > 200) BRUIT = 200; break;
    case 'e': EPSILON = std::max(EPSILON - 1, 0);   Net.set_epsilon(EPSILON); break;
    case 'E': EPSILON = std::min(EPSILON + 1, 200); Net.set_epsilon(EPSILON); break;
    case 'n': normalized = !normalized; break;
  }
}

/*===========================================================================*\
 * update_active_fonts                                                       *
\*===========================================================================*/
static void update_active_fonts()
{
  active_fonts.clear();
  for (unsigned i = 0; i < all_fonts.size(); i++)
    if (all_fonts[i].active)
      active_fonts.push_back("fonts/" + all_fonts[i].name + ".ttf");
}

/*===========================================================================*\
 * special_callback                                                          *
\*===========================================================================*/
static void special_callback(int key, int, int)
{
  switch (key)
  {
    case GLUT_KEY_F1:
      if (all_fonts.size() > 0)
        all_fonts[0].active = !all_fonts[0].active;
      break;
    case GLUT_KEY_F2:
      if (all_fonts.size() > 1)
        all_fonts[1].active = !all_fonts[1].active;
      break;
    case GLUT_KEY_F3:
      if (all_fonts.size() > 2)
        all_fonts[2].active = !all_fonts[2].active;
      break;
    case GLUT_KEY_F4:
      if (all_fonts.size() > 3)
        all_fonts[3].active = !all_fonts[3].active;
      break;
    case GLUT_KEY_F5:
      if (all_fonts.size() > 4)
        all_fonts[4].active = !all_fonts[4].active;
      break;
    case GLUT_KEY_F6:
      if (all_fonts.size() > 5)
        all_fonts[5].active = !all_fonts[5].active;
      break;
    case GLUT_KEY_F7:
      if (all_fonts.size() > 6)
        all_fonts[6].active = !all_fonts[6].active;
      break;
    case GLUT_KEY_F8:
      if (all_fonts.size() > 7)
        all_fonts[7].active = !all_fonts[7].active;
      break;
    case GLUT_KEY_F9:
      if (all_fonts.size() > 8)
        all_fonts[8].active = !all_fonts[8].active;
      break;
    case GLUT_KEY_F10:
      if (all_fonts.size() > 9)
        all_fonts[9].active = !all_fonts[9].active;
      break;
    case GLUT_KEY_F11:
      if (all_fonts.size() > 10)
        all_fonts[10].active = !all_fonts[10].active;
      break;
    case GLUT_KEY_F12:
      if (all_fonts.size() > 11)
        all_fonts[11].active = !all_fonts[11].active;
      break;
    case GLUT_KEY_LEFT:
      min_font_size = std::max(min_font_size - 1, 8);
      break;
    case GLUT_KEY_RIGHT:
      min_font_size = std::min(min_font_size + 1, max_font_size);
      break;
    case GLUT_KEY_UP:
      max_font_size = std::min(max_font_size + 1, taille_image);
      break;
    case GLUT_KEY_DOWN:
      max_font_size = std::max(max_font_size - 1, min_font_size);
      break;
  }
  update_active_fonts();
  if (active_fonts.empty())
  {
    all_fonts[0].active = true;
    update_active_fonts();
  }
}

/*===========================================================================*\
 * split                                                                     *
\*===========================================================================*/
static void split(const std::string &s, char delim, std::vector<std::string> &elems)
{
  std::stringstream ss(s);
  std::string item;
  while (std::getline(ss, item, delim))
    elems.push_back(item);
}

/*===========================================================================*\
 * parse_arguments                                                           *
\*===========================================================================*/
static void parse_arguments(char *argv[])
{
  for (int i = 1; argv[i]; i++)
  {
    std::string arg(argv[i]);
    std::vector<std::string> elems;
    split(arg, '=', elems);
    if (elems.size() != 2)
    {
      std::cerr << "Argument inconnu : '" << arg << "' (ignoré)\n";
      continue;
    }
    if (elems[0] == "bruit")
    {
      int value = -1;
      std::stringstream ss(elems[1]);
      ss >> value;
      if (value >= 0 && value <= 200)
      {
        BRUIT = value;
        std::cout << "Valeur de bruit utilisée : " << BRUIT << "%\n";
      }
      else
        std::cerr << "Mauvaise valeur de bruit : " << elems[1] << "\n";
    } else if (elems[0] == "epsilon")
    {
      int value = -1;
      std::stringstream ss(elems[1]);
      ss >> value;
      if (value >= 0 && value <= 200)
      {
        EPSILON = value;
        std::cout << "Valeur d'epsilon utilisée : " << EPSILON << "%\n";
      }
      else
        std::cerr << "Mauvaise valeur d'epsilon : " << elems[1] << "\n";
    } else if (elems[0] == "polices")
    {
      std::vector<std::string> polices;
      split(elems[1], ',', polices);
      for (auto &police : polices)
      {
        if (access(("fonts/" + police + ".ttf").c_str(), R_OK) == 0)
        {
          all_fonts.push_back(Font(police));
          std::cout << "Ajout de la police : " << police << std::endl;
        } else
          std::cerr << "Police introuvable : '" << police << "'\n";
      }
    } else if (elems[0] == "taille_police")
    {
      
    } else if (elems[0] == "taille_image")
    {
      int value = -1;
      std::stringstream ss(elems[1]);
      ss >> value;
      if (value >= 16 && value <= 64)
      {
        BRUIT = value;
        std::cout << "Valeur de bruit utilisée: " << BRUIT << "%\n";
      }
      else
      {
        std::cerr << "Mauvaise valeur de bruit: " << elems[1] << "\n";
      }
    }
    else
      std::cerr << "Argument inconnu: '" << arg << "' (ignoré)\n";
  }
}

/*****************************************************************************\
 * main                                                                      *
 * point d'entrée du programme                                               *
\*****************************************************************************/
int main(int argc, char *argv[])
{
  srand(getpid()+ time(0)); // Initialisation des nombres aléatoires

  all_fonts.push_back(Font("luxisr"));
  parse_arguments(argv);
  update_active_fonts();

  Net.init(EPSILON, neurones_par_couche);       // initialisation du réseau de neurones

  // Initialisation graphique
  glutInitContextVersion(3, 3);
  glutInitContextFlags(GLUT_FORWARD_COMPATIBLE | GLUT_DEBUG);
  glutInitContextProfile(GLUT_CORE_PROFILE);
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGB);
  glutInitWindowSize(1000,63*6-15);
  glutCreateWindow("Réseau de Neurones");
  glutDisplayFunc(display_callback);
  glutKeyboardFunc(keyboard_callback);
  glutSpecialFunc(special_callback);
  glutIdleFunc(idle_callback);
 
  glewExperimental=true;glewInit();glGetError();

  FONT_MANAGER.init();
  FONT_MANAGER.set_screen_size(vec2(1000, 63*6-15));
  FONT_MANAGER.set_font("fonts/luxisr.ttf", 8);
  
  PRINT_OPENGL_ERROR();

  rectangles_init();
  init_opengl();

  glutMainLoop();

  return 0;
}
